@@ -6,10 +6,10 @@
 
 #include "speller.hpp"
 #include "convert.hpp"
-#include "clone_ptr-t.hpp"
+#include "clone_ptr.hpp"
 #include "config.hpp"
 
-namespace acommon {
+namespace aspell {
 
   Speller::Speller(SpellerLtHandle h) : lt_handle_(h) {}
 
