@@ -7,16 +7,15 @@
 #include "enumeration.hpp"
 #include "parm_string.hpp"
 
-using namespace acommon;
 
-namespace aspeller {
+namespace aspell { namespace sp {
 
   class SpellerImpl;
 
   class SuggestionList : public WordList {
   public:
-    typedef StringEnumeration            VirEmul;
-    typedef Enumeration<VirEmul>         Emul;
+    typedef StringEnumeration          VirEmul;
+    typedef Enumeration<VirEmul>       Emul;
     typedef const char *               Value;
     typedef unsigned int               Size;
 
@@ -32,12 +31,13 @@ namespace aspeller {
   class Suggest {
   public:
     virtual PosibErr<void> set_mode(ParmString) = 0;
+    virtual double score(const char * base, const char * other) = 0;
     virtual SuggestionList & suggest(const char * word) = 0;
     virtual ~Suggest() {}
   };
   
   PosibErr<Suggest *> new_default_suggest(SpellerImpl *);
-}
+} }
 
 
 #endif
